@@ -3,15 +3,18 @@ import gzip
 import json
 import logging
 import time
+from asyncio import sleep
 from enum import Enum
 from random import random
+from socket import gaierror
 from typing import Optional, List, Dict, Callable, Any
 
 import websockets as ws
+from websockets.exceptions import ConnectionClosedError
 
 from .client import AsyncClient
-from .exceptions import BinanceWebsocketUnableToConnect
 from .enums import FuturesType
+from .exceptions import BinanceWebsocketUnableToConnect
 from .enums import ContractType
 from .threaded_stream import ThreadedApiManager
 
@@ -34,12 +37,12 @@ class BinanceSocketType(str, Enum):
 
 
 class ReconnectingWebsocket:
-
     MAX_RECONNECTS = 5
     MAX_RECONNECT_SECONDS = 60
     MIN_RECONNECT_WAIT = 0.1
     TIMEOUT = 10
     NO_MESSAGE_RECONNECT_TIMEOUT = 60
+    MAX_QUEUE_SIZE = 100
 
     def __init__(
         self, loop, url: str, path: Optional[str] = None, prefix: str = 'ws/', is_binary: bool = False, exit_coro=None
@@ -57,6 +60,7 @@ class ReconnectingWebsocket:
         self.ws: Optional[ws.WebSocketClientProtocol] = None
         self.ws_state = WSListenerState.INITIALISING
         self._queue = asyncio.Queue(loop=self._loop)
+        self._handle_read_loop = None
 
     async def __aenter__(self):
         await self.connect()
@@ -68,24 +72,34 @@ class ReconnectingWebsocket:
         self.ws_state = WSListenerState.EXITING
         if self.ws:
             self.ws.fail_connection()
-        if self._conn:
+        if self._conn and hasattr(self._conn, 'protocol'):
             await self._conn.__aexit__(exc_type, exc_val, exc_tb)
         self.ws = None
+        if not self._handle_read_loop:
+            self._log.error("CANCEL read_loop")
+            await self._kill_read_loop()
 
     async def connect(self):
         await self._before_connect()
         assert self._path
+        self.ws_state = WSListenerState.STREAMING
         ws_url = self._url + self._prefix + self._path
-        self._conn = ws.connect(ws_url, close_timeout=0.001)
+        self._conn = ws.connect(ws_url, close_timeout=0.1)
         try:
             self.ws = await self._conn.__aenter__()
         except:  # noqa
             await self._reconnect()
             return
-        self.ws_state = WSListenerState.STREAMING
         self._reconnects = 0
         await self._after_connect()
-        self._loop.call_soon_threadsafe(asyncio.create_task, self._read_loop())
+        # To manage the "cannot call recv while another coroutine is already waiting for the next message"
+        if not self._handle_read_loop:
+            self._handle_read_loop = self._loop.call_soon_threadsafe(asyncio.create_task, self._read_loop())
+
+    async def _kill_read_loop(self):
+        self.ws_state = WSListenerState.EXITING
+        while self._handle_read_loop:
+            await sleep(0.1)
 
     async def _before_connect(self):
         pass
@@ -106,46 +120,75 @@ class ReconnectingWebsocket:
             return None
 
     async def _read_loop(self):
-        while True:
-            res = None
-            if not self.ws or self.ws_state != WSListenerState.STREAMING:
-                await self._wait_for_reconnect()
-                break
-            if self.ws_state == WSListenerState.EXITING:
-                break
-            if self.ws.state == ws.protocol.State.CLOSING:
-                break
-            if self.ws.state == ws.protocol.State.CLOSED:
+        try:
+            while True:
                 try:
-                    await self._reconnect()
-                except BinanceWebsocketUnableToConnect:
-                    return {
-                        'e': 'error',
-                        'm': 'Max reconnect retries reached'
-                    }
-                else:
-                    break
-            try:
-                res = await asyncio.wait_for(self.ws.recv(), timeout=self.TIMEOUT)
-            except asyncio.TimeoutError:
-                self._log.debug(f"no message in {self.TIMEOUT} seconds")
-            except asyncio.CancelledError as e:
-                self._log.debug(f"cancelled error {e}")
-                break
-            except asyncio.IncompleteReadError as e:
-                self._log.debug(f"incomplete read error {e}")
-            except Exception as e:
-                self._log.debug(f"exception {e}")
-                break
-            else:
-                if self.ws_state in (WSListenerState.EXITING, WSListenerState.RECONNECTING):
-                    break
-                res = self._handle_message(res)
-                if self.ws_state in (WSListenerState.EXITING, WSListenerState.RECONNECTING):
-                    break
+                    if self.ws_state == WSListenerState.RECONNECTING:
+                        await self._run_reconnect()
 
-            if res and self._queue.qsize() < 100:
-                await self._queue.put(res)
+                    if not self.ws or self.ws_state != WSListenerState.STREAMING:
+                        await self._wait_for_reconnect()
+                        break
+                    elif self.ws_state == WSListenerState.EXITING:
+                        break
+                    elif self.ws.state == ws.protocol.State.CLOSING:
+                        await asyncio.sleep(0.1)
+                        continue
+                    elif self.ws.state == ws.protocol.State.CLOSED:
+                        await self._reconnect()
+                    elif self.ws_state == WSListenerState.STREAMING:
+                        res = await asyncio.wait_for(self.ws.recv(), timeout=self.TIMEOUT)
+                        res = self._handle_message(res)
+                        if res:
+                            if self._queue.qsize() < self.MAX_QUEUE_SIZE:
+                                await self._queue.put(res)
+                            else:
+                                self._log.debug(f"Queue overflow {self.MAX_QUEUE_SIZE}. Message not filled")
+                                await self._queue.put({
+                                    'e': 'error',
+                                    'm': 'Queue overflow. Message not filled'
+                                })
+                                raise BinanceWebsocketUnableToConnect
+                except asyncio.TimeoutError:
+                    self._log.debug(f"no message in {self.TIMEOUT} seconds")
+                    # _no_message_received_reconnect
+                except asyncio.CancelledError as e:
+                    self._log.debug(f"cancelled error {e}")
+                    break
+                except asyncio.IncompleteReadError as e:
+                    self._log.debug(f"incomplete read error ({e})")
+                except ConnectionClosedError as e:
+                    self._log.debug(f"connection close error ({e})")
+                except gaierror as e:
+                    self._log.debug(f"DNS Error ({e})")
+                except BinanceWebsocketUnableToConnect as e:
+                    self._log.debug(f"BinanceWebsocketUnableToConnect ({e})")
+                    break
+                except Exception as e:
+                    self._log.debug(f"Unknown exception ({e})")
+                    continue
+        finally:
+            self._handle_read_loop = None  # Signal the coro is stopped
+            self._reconnects = 0
+
+    async def _run_reconnect(self):
+        await self.before_reconnect()
+        if self._reconnects < self.MAX_RECONNECTS:
+            reconnect_wait = self._get_reconnect_wait(self._reconnects)
+            self._log.debug(
+                f"websocket reconnecting. {self.MAX_RECONNECTS - self._reconnects} reconnects left - "
+                f"waiting {reconnect_wait}"
+            )
+            await asyncio.sleep(reconnect_wait)
+            await self.connect()
+        else:
+            self._log.error(f'Max reconnections {self.MAX_RECONNECTS} reached:')
+            # Signal the error
+            await self._queue.put({
+                'e': 'error',
+                'm': 'Max reconnect retries reached'
+            })
+            raise BinanceWebsocketUnableToConnect
 
     async def recv(self):
         res = None
@@ -157,12 +200,8 @@ class ReconnectingWebsocket:
         return res
 
     async def _wait_for_reconnect(self):
-        while self.ws_state == WSListenerState.RECONNECTING:
-            self._log.debug("reconnecting waiting for connect")
-        if not self.ws:
-            self._log.debug("ignore message no ws")
-        else:
-            self._log.debug(f"ignore message {self.ws_state}")
+        while self.ws_state != WSListenerState.STREAMING:
+            await sleep(0.1)
 
     def _get_reconnect_wait(self, attempts: int) -> int:
         expo = 2 ** attempts
@@ -176,24 +215,10 @@ class ReconnectingWebsocket:
 
     def _no_message_received_reconnect(self):
         self._log.debug('No message received, reconnecting')
-        asyncio.create_task(self._reconnect())
+        self.ws_state = WSListenerState.RECONNECTING
 
     async def _reconnect(self):
-        if self.ws_state == WSListenerState.RECONNECTING:
-            return
         self.ws_state = WSListenerState.RECONNECTING
-        await self.before_reconnect()
-        if self._reconnects < self.MAX_RECONNECTS:
-            reconnect_wait = self._get_reconnect_wait(self._reconnects)
-            self._log.debug(
-                f"websocket reconnecting {self.MAX_RECONNECTS - self._reconnects} reconnects left - "
-                f"waiting {reconnect_wait}"
-            )
-            await asyncio.sleep(reconnect_wait)
-            await self.connect()
-        else:
-            self._log.error(f'Max reconnections {self.MAX_RECONNECTS} reached:')
-            raise BinanceWebsocketUnableToConnect
 
 
 class KeepAliveWebsocket(ReconnectingWebsocket):
@@ -245,30 +270,32 @@ class KeepAliveWebsocket(ReconnectingWebsocket):
         return listen_key
 
     async def _keepalive_socket(self):
-        listen_key = await self._get_listen_key()
-
-        if listen_key != self._path:
-            self._log.debug("listen key changed: reconnect")
-            self._path = listen_key
-            await self._reconnect()
-        else:
-            self._log.debug("listen key same: keepalive")
-            if self._keepalive_type == 'user':
-                await self._client.stream_keepalive(self._path)
-            elif self._keepalive_type == 'margin':  # cross-margin
-                await self._client.margin_stream_keepalive(self._path)
-            elif self._keepalive_type == 'futures':
-                await self._client.futures_stream_keepalive(self._path)
-            elif self._keepalive_type == 'coin_futures':
-                await self._client.futures_coin_stream_keepalive(self._path)
-            else:  # isolated margin
-                # Passing symbol for isolated margin
-                await self._client.isolated_margin_stream_keepalive(self._keepalive_type, self._path)
+        try:
+            listen_key = await self._get_listen_key()
+            if listen_key != self._path:
+                self._log.debug("listen key changed: reconnect")
+                self._path = listen_key
+                await self._reconnect()
+            else:
+                self._log.debug("listen key same: keepalive")
+                if self._keepalive_type == 'user':
+                    await self._client.stream_keepalive(self._path)
+                elif self._keepalive_type == 'margin':  # cross-margin
+                    await self._client.margin_stream_keepalive(self._path)
+                elif self._keepalive_type == 'futures':
+                    await self._client.futures_stream_keepalive(self._path)
+                elif self._keepalive_type == 'coin_futures':
+                    await self._client.futures_coin_stream_keepalive(self._path)
+                else:  # isolated margin
+                    # Passing symbol for isolated margin
+                    await self._client.isolated_margin_stream_keepalive(self._keepalive_type, self._path)
+        except Exception:
+            pass  # Ignore
+        finally:
             self._start_socket_timer()
 
 
 class BinanceSocketManager:
-
     STREAM_URL = 'wss://stream.binance.{}:9443/'
     STREAM_TESTNET_URL = 'wss://testnet.binance.vision/'
     FSTREAM_URL = 'wss://fstream.binance.{}/'
@@ -531,8 +558,8 @@ class BinanceSocketManager:
             <pair>_<contractType>@continuousKline_<interval>
         """
 
-        path = f'{symbol.lower()}_{contract_type}@continuousKline_{interval}'
-        return self._get_futures_socket(path, futures_type=futures_type)
+        path = f'{symbol.lower()}_{contract_type.value}@continuousKline_{interval}'
+        return self._get_futures_socket(path, prefix='ws/', futures_type=futures_type)
 
     def miniticker_socket(self, update_time: int = 1000):
         """Start a miniticker websocket for all trades
@@ -1143,7 +1170,7 @@ class BinanceSocketManager:
         if conn_key not in self._conns:
             return
 
-        del(self._conns[conn_key])
+        del (self._conns[conn_key])
 
 
 class ThreadedWebsocketManager(ThreadedApiManager):
@@ -1204,8 +1231,8 @@ class ThreadedWebsocketManager(ThreadedApiManager):
             params={
                 'symbol': symbol,
                 'interval': interval,
-                'futures_type': futures_type.value,
-                'contract_type': contract_type.value
+                'futures_type': futures_type,
+                'contract_type': contract_type
             }
         )
 
